
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GatherSkeleton'                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class GatherSkeleton : public DataSchedule {
public :
  GatherSkeleton(const Frame* _loc, const Frame* _rem, const int len,
                 const Frame**_subs, int** subsData,
                 const Frame* _msk, int* mskData) ;

private :
  const Frame* loc ;
  const Frame* rem ;
  const Frame** subs ;
  const Frame*  msk ;

  int rank ;

  void loop(int locOffset, int** subsData, int* mskData, const int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapeSkeleton'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ReshapeSkeleton : public DataSchedule {
public :
  ReshapeSkeleton(const Frame* _loc, const Frame* _rem, const int len) ;
private :
  const Frame* loc ;
  const Frame* rem ;

  int rank ;

  void loop(int locOffset, int seqIndex, int seqStride, int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapePadSkeleton'                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ReshapePadSkeleton : public DataSchedule {
public :
  ReshapePadSkeleton(const Frame* _loc, int _remSize, const Frame* _pad,
                     const int len) ;
private :
  const Frame* loc ;
  const Frame* pad ;

  int rank ;
  int remSize ;

  void loop(int locOffset, int seqIndex, int seqStride, int r) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherSkeleton'                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecGatherSkeleton : public BlockDataSchedule {
public :
  VecGatherSkeleton(const Frame* _loc, const Frame* _rem, const int len,
                    const Frame** _subs, int** _subsData) ;

protected :

  int* exts ;
  int* locSteps ;
  int* remSteps ;

private :
  const Frame* loc ;
  const Frame* rem ;
  const Frame** subs ;
  int** subsData ;

  int rank ;

  static int blockRank(const int rank, const Frame** subs) ;

  void loop(const int locOffset, const int locCtg,
            const int remOffset, const int remCtg,
            const int vol,
            Group remGrp,
            const int r, const int s) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Gather'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Gather {
public :
  Gather(const Frame* loc, const Frame* rem, const int len,
         const Frame** subs, int** subsData,
         const Frame* msk, int* mskData) :
      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(loc, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.gather(locDat, remDat) ;
    bcst.execute(locDat) ;
  }
private :
  GatherSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Scatter'                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Scatter {
public :
  Scatter(const Frame* loc, const Frame* rem, const int len,
          const Frame** subs, int** subsData,
          const Frame* msk, int* mskData) :
      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(loc, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.scatter(locDat, remDat) ;
    bcst.execute(locDat) ;
  }
private :
  GatherSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ScatterComb'                                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class ScatterComb {
public :
  ScatterComb(const Frame* loc, const Frame* rem, const int len,
              const Frame** subs, int** subsData,
              const Frame* msk, int* mskData,
              void (*_op)(void*, void*)) :
      skeleton(loc, rem, len, subs, subsData, msk, mskData), bcst(rem, len),
      op(_op) {}

  void execute(void* locDat, void* remDat) {
    skeleton.combine(locDat, remDat, op) ;
    bcst.execute(locDat) ;
  }
private :
  GatherSkeleton skeleton ;
  Replicate bcst ;
  void (*op)(void*, void*) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Reshape'                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Reshape {
public :
  Reshape(const Frame* loc, const Frame* rem, const int len) :
              skeleton(loc, rem, len), bcst(loc, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.gather(locDat, remDat) ;
    bcst.execute(locDat) ;
  }
protected :
  ReshapeSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapePad'                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ReshapePad {
public :
  ReshapePad(const Frame* loc, int remSize, const Frame* pad, const int len) :
      skeleton(loc, remSize, pad, len), bcst(loc, len) {}

  void execute(void* locDat, void* padDat) {
    skeleton.gather(locDat, padDat) ;
    bcst.execute(padDat) ;
  }
protected :
  ReshapePadSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGather'                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecGather {
public :
  VecGather(const Frame* loc, const Frame* rem, const int len,
            const Frame** subs, int** subsData) :
      skeleton(loc, rem, len, subs, subsData), bcst(loc, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.gather(locDat, remDat) ;
    bcst.execute(locDat) ;
  }
private :
  VecGatherSkeleton skeleton ;
  Replicate bcst ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecScatter'                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class VecScatter {
public :
  VecScatter(const Frame* loc, const Frame* rem, const int len,
             const Frame** subs, int** subsData) :
      skeleton(loc, rem, len, subs, subsData), bcst(rem, len) {}

  void execute(void* locDat, void* remDat) {
    skeleton.scatter(locDat, remDat) ;
    bcst.execute(remDat) ;
  }
private :
  VecGatherSkeleton skeleton ;
  Replicate bcst ;
} ;

